#!/usr/bin/perl
#
# written by Assen Roguev, 2006
# questions on the code to: roguev@cmp.ucsf.edu or stoynov@obzor.bio21.bas.bg
#
# thermo_DD.pl
# usage:
#		./thermo_DD.pl input_file start_position stop_position window_size step > output_file
# 
# calculates the thermodynamic stability of DNA-DNA duplexes
# for given set of nearest-neighbour parameters using a sliding window with arbitrtary size and step
# The values of the nearest-neighbor interactions for the DNA/DNA duplex are reported by Breslauer & Marky (1986)Proc Natl Acad Sci U S A, 83, 3746-3750.
# 
# this software is freeware
# feel free to distribute and modify it
#

# let the code begin
use strict;

# hash of arrays with thermodynamic parameters
# hash keys are respective pairs
# first array element is enthalpy (dH)
# second array element is enthropy (dS)
my %data_hash = (
'AA' => [9.1, 0.024],
'TT' => [9.1, 0.024],
'AT' => [8.6, 0.0239],
'TA' => [6.0, 0.0169],
'CA' => [5.8, 0.0129],
'TG' => [5.8, 0.0129],
'GT' => [6.5, 0.0173],
'AC' => [6.5, 0.0173],
'CT' => [7.8, 0.0208],
'AG' => [7.8, 0.0208],
'GA' => [5.6, 0.0135],
'TC' => [5.6, 0.0135],
'CG' => [11.9, 0.0278],
'GC' => [11.1, 0.0267],
'GG' => [11.0, 0.0266],
'CC' => [11.0, 0.0266]
);

# input parameters for the script
my $seq_file = shift;			# sequence file
my $start = shift;				# calculation start position 
my $stop = shift;				# calculation end position
my $window = shift;				# window size
my $step = shift;				# step

# some variables we are gonna use later
# define them here as globals to avoid scope issues
my $S;				# entropy
my $H;				# entalpy
my $G;				# free energy
my $roundG;			# rounded free energy, used for the output
my $input_seq;		# input sequence after stiching the lines	
my $current_pos;	# current position within the sequence
my $seq;		# working sequence
my $N_character = 'N';	# N unidentified nucleotide in the sequence

# open the input file
open(IN,$seq_file) || die "Cannot open $seq_file\n";

# read line by line
while (my $line = <IN>)					# read a line
	{
	chomp($line);						# remove new line characters at the end of the lines
	$input_seq = $input_seq.$line;		# stitch the line with the rest
	}
close(IN);								# close input file

# define the working sequence based on start and stop position 
$seq = substr($input_seq, $start, ($stop-$start)); 

# do the maths
for (my $i = 0; $i <= length($seq) - $window; $i+= $step)	# incerement by the given step
	{
	# zero the variables
	$S = 0;
	$H = 0;
	$G = 0;

	# define a subsequnce based on the window size 
	my $sub_seq = substr($seq,$i,$window);
	
	# removes sequences with Ns in them
	if ($sub_seq =~ /$N_character/) { next; }

	# extract all the nearest neighbour pairs
	for (my $j = 0; $j < length($sub_seq); $j++)
		{
		my $pair = substr($sub_seq,$j,2);
	
		$pair = uc($pair);		# convert to upper case
		
		# get respective values from the hash
		$H += $data_hash{$pair}->[0];
		$S += $data_hash{$pair}->[1]; 
		
		# calculate free energy
		$G =$H*(1-310/($H/$S-36)); 
		}
		
	$roundG = sprintf "%.2f",$G;	# round the value	
	$current_pos = $start + 1 + $i + $window/2;		# calculate where we are now. If window size is an odd number, add 0.5
	print "$current_pos\t$roundG\n";				# print output on screen
	}